package name.panitz.ki
case class Atom(name:String) extends Formel{
  override def toString()=name}
case class Not(e1:Formel) extends Formel{
  override def toString()="¬"+e1.toString}
case class Impl(e1:Formel,e2:Formel) extends Formel{
  override def toString()="("+e1.toString+" -> "+e2.toString+")"}
case class And(e1:Formel,e2:Formel) extends Formel{
  override def toString()="("+e1.toString+" /\\ "+e2.toString+")"}
case class Or(e1:Formel,e2:Formel) extends Formel{
  override def toString()="("+e1.toString+" \\/ "+e2.toString+")"}

abstract class Formel{
  def eliminateImpl():Formel={
   this match {
    case Atom(_)  => this
    case Not(x)   => Not(x.eliminateImpl)
    case And(x,y) => And(x.eliminateImpl,y.eliminateImpl)
    case Or(x,y)  => Or(x.eliminateImpl,y.eliminateImpl)
    case Impl(x,y)=> Or(Not(x.eliminateImpl),y.eliminateImpl)
   }
  }

  def negationToInner():Formel={
   this match {
    case Atom(_)       => this
    case Not(Atom(x))  => this
    case Not(Not(x))   => x.negationToInner
    case Not(And(x,y)) => Or(Not(x),Not(y)).negationToInner
    case Not(Or(x,y))  => And(Not(x),Not(y)).negationToInner
    case And(x,y)      => And(x.negationToInner,y.negationToInner)
    case Or(x,y)       => Or(x.negationToInner,y.negationToInner)
   }
  }

  def makeConjunctions():Formel={
   this match {
    case Atom(_)       => this
    case Not(x)        => this
    case Or(And(x,y),z)=> And(Or(x,z),Or(y,z)).makeConjunctions
    case Or(x,And(y,z))=> And(Or(x,y),Or(x,z)).makeConjunctions
    case Or(x,y) 
      => val x1 =x.makeConjunctions
         val y1 =y.makeConjunctions
         x1 match { 
           case And(_,_) => Or(x1,y1).makeConjunctions
           case _        => y1 match {
                           case And(_,_) => Or(x1,y1).makeConjunctions
                           case _        => Or(x1,y1)
	                 }
         }
    case And(x,y)      => And(x.makeConjunctions,y.makeConjunctions)
   }
  }

  def toCNF():Formel=eliminateImpl.negationToInner.makeConjunctions

  import scala.collection.mutable.HashSet
  import scala.collection.mutable.Set
  def conjunctionSet():Set[Formel]={
    val result=new HashSet[Formel]()
    this match {
      case And(e1,e2) => result++= e1.conjunctionSet;
                         result++= e2.conjunctionSet;
      case e => result+= e
    }
    result
  }

  type Klausel=Set[Formel]
  def disjunctionSet():Klausel={
    val result=new HashSet[Formel]()
    this match {
      case Or(e1,e2) => result++= e1.disjunctionSet;
                        result++= e2.disjunctionSet;
      case e => result+= e
    }
    result
  }

  def cnf():List[Klausel]=
    (for (val f<- toCNF.conjunctionSet.elements) 
       yield f.disjunctionSet).toList

  def negated()= this match { case Not(x) => x
                              case y      => Not(y)}

}

