package name.panitz.ki
class Scene extends scala.collection.mutable.ArrayBuffer[Vertex]
            with    DepthSearch[CaseClass] {

  def this(that:Scene)={
    this()
    import scala.collection.mutable.HashMap;
    val copiedEdges=new HashMap[Edge,Edge]()
    def cop(e:Edge)={
      try {copiedEdges(e)}
      catch{case _:java.lang.Error=>
        val copy=new Edge(e.marked)
        copiedEdges.update(e,copy)
        copy
      }
    }
    for (val v<-that.elements) {
      v match {
        case L(e1,e2)=>this += L(cop(e1),cop(e2))
        case T(e1,e2,e3)=>this += T(cop(e1),cop(e2),cop(e3))
        case Fork(e1,e2,e3)=>this += Fork(cop(e1),cop(e2),cop(e3))
        case Arrow(e1,e2,e3)=>this += Arrow(cop(e1),cop(e2),cop(e3))
      }
    }
  }


  def possibleMarks(v:Vertex)={
    import Marked.{_};
    v match{
      case L(_,_)
        => List(Tuple(BoundaryLeft,BoundaryLeft)
               ,Tuple(BoundaryRight,BoundaryRight)
               ,Tuple(BoundaryRight,Convex)
               ,Tuple(Convex,BoundaryRight)
               ,Tuple(BoundaryLeft,Concave)
               ,Tuple(Concave,BoundaryLeft)
               )
      case Fork(_,_,_)
        => List(Tuple(Convex,Convex,Convex)
               ,Tuple(Concave,Concave,Concave)
               ,Tuple(BoundaryLeft,Concave,BoundaryLeft)
               ,Tuple(BoundaryLeft,BoundaryLeft,Concave)
               ,Tuple(Concave,BoundaryLeft,BoundaryLeft)
               )
      case T(_,_,_)
       => List(Tuple(BoundaryLeft,BoundaryLeft,BoundaryLeft)
              ,Tuple(BoundaryLeft,BoundaryRight,BoundaryLeft)
              ,Tuple(BoundaryLeft,Concave,BoundaryLeft)
              ,Tuple(BoundaryLeft,Convex,BoundaryLeft)
              )
      case Arrow(_,_,_)
       => List(Tuple(BoundaryLeft,Convex,BoundaryLeft)
              ,Tuple(Convex,Concave,Convex)
              ,Tuple(Concave,Convex,Concave)
              )
    }
  }

  def correctlyMarked(v:Vertex)
    =possibleMarks(v).exists((p)=>markFits(v,p))

  def markFits(v:Vertex,t:CaseClass)=
    Pair(v,t) match{
      case Pair(L(e1,e2),Tuple2(m1,m2)) => 
        fit(e1.marked,m1)&&fit(e2.marked,m2)
      case Pair(T(e1,e2,e3),Tuple3(m1,m2,m3)) =>
        fit(e1.marked,m1)&&fit(e2.marked,m2)&&fit(e3.marked,m3)
      case Pair(Fork(e1,e2,e3),Tuple3(m1,m2,m3)) => 
        fit(e1.marked,m1)&&fit(e2.marked,m2)&&fit(e3.marked,m3)
      case Pair(Arrow(e1,e2,e3),Tuple3(m1,m2,m3)) => 
        fit(e1.marked,m1)&&fit(e2.marked,m2)&&fit(e3.marked,m3)
    }

  def fit(x:Marked.Value,y:Any)=null==x||x==y

  def moves():List[CaseClass]={
    val unmarkedVertexes
      = for (val v<-this.elements;!isMarked(v)) yield v

    if (unmarkedVertexes.hasNext){
      val v=unmarkedVertexes.next
      (for (val pm<-possibleMarks(v)
           ;markFits(v,pm)
           ;this.elements.forall((x)=>correctlyMarked(x))
           ) yield pm).toList
    }else List()
  } 

  def isMarked(v:Vertex)=
    v match{
      case L(e1,e2) =>e1.marked!=null&&e2.marked!=null
      case T(e1,e2,e3) 
        =>e1.marked!=null&&e2.marked!=null&&e3.marked!=null
      case Fork(e1,e2,e3) 
        =>e1.marked!=null&&e2.marked!=null&&e3.marked!=null
      case Arrow(e1,e2,e3)
        =>e1.marked!=null&&e2.marked!=null&&e3.marked!=null
    } 

  def move(m:CaseClass):Scene={
    val Pair(marked,unmarked)
     =new Scene(this).toList.span((v)=>isMarked(v))
    val result=new Scene()    
    for (val v<-marked) result += v
    result += markedCopy(unmarked.head,m)
    for (val v<-unmarked.tail) result += v
    result
  }

  def markedCopy(v:Vertex,t:CaseClass)={
    def mv(e:Edge,m:Any)={e.marked=m.asInstanceOf[Marked.Value];e}
    Pair(v,t) match{
      case Pair(L(e1,e2),Tuple2(m1,m2)) 
       => L(mv(e1,m1),mv(e2,m2))
      case Pair(T(e1,e2,e3),Tuple3(m1,m2,m3)) 
       => T(mv(e1,m1),mv(e2,m2),mv(e3,m3))
      case Pair(Fork(e1,e2,e3),Tuple3(m1,m2,m3)) 
       => Fork(mv(e1,m1),mv(e2,m2),mv(e3,m3))
      case Pair(Arrow(e1,e2,e3),Tuple3(m1,m2,m3)) 
       => Arrow(mv(e1,m1),mv(e2,m2),mv(e3,m3))
    }
  }

  def terminalState():Boolean=
   elements.forall((v)=>isMarked(v))
}

